{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at
http://www.mozilla.org/NPL/NPL-1_1Final.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: mwGenericHashList.pas, released January 28, 2001.

The Initial Developer of the Original Code is Martin Waldenburg
(Martin.Waldenburg@T-Online.de).
Portions created by Martin Waldenburg are Copyright (C) 2000, 2001 Martin Waldenburg.
All Rights Reserved.

Contributor(s): ___________________.

Last Modified: 3/7/2001
Current Version: 1.2

Notes: This is a very fast generic Hash list.
       As any kind of data can be represented
       as a string you can use it to hash any
       imageable kind of data.
       Just derive from TmwBaseHashWord and
       assign your Hash data in the constructor
       to "S".

Known Issues:
-----------------------------------------------------------------------------}
{$R-}
unit mwGenericHashList;

interface

uses Classes, SysUtils, Windows;

var
  mwHashTable: array[#0..#255] of byte;
  mwInsensitiveHashTable: array[#0..#255] of Byte;

type
  TmwStringHash = function(const aString: string): Integer;
  TmwStringHashCompare = function(const Str1: string; const Str2: string): Boolean;

  TmwBaseHashWord = class
    S: string;
  end;

  PHashPointerList = ^THashPointerList;
  THashPointerList = array[1..1] of TObject;

  TmwBaseGenericHashList = class(TObject)
    FList: PHashPointerList;
    fCapacity: Integer;
  protected
    fCompare: TmwStringHashCompare;
    fHash: TmwStringHash;
    procedure AdjustCapacity;
    function Get(Index: Integer): Pointer;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
  public
    constructor Create(aCompare: TmwStringHashCompare);
    destructor Destroy; override;
    procedure Clear;
    property Capacity: Integer read fCapacity;
    property Items[Index: Integer]: Pointer read Get write Put; default;
  end;

  TmwGenericHashStrings = class(TmwBaseGenericHashList)
  public
    function delete(const aString: string): Integer;
    procedure Add(const aHashWord: TmwBaseHashWord);
  end;

  TmwGenericHashItems = class(TmwBaseGenericHashList)
  public
    constructor Create(aHash: TmwStringHash; aCompare: TmwStringHashCompare);
    procedure Add(const aHashWord: TmwBaseHashWord);
    function delete(const aString: string): Integer;
    function Hash(const S: string): Pointer;
    function HashEX(const S: string; HashValue: Integer): Pointer;
  end;

  TmwGenericHashList = class(TmwBaseGenericHashList)
  private
    fSecondaryHash: TmwStringHash;
  public
    constructor Create(Primary, Secondary: TmwStringHash; aCompare: TmwStringHashCompare);
    procedure Add(const aHashWord: TmwBaseHashWord);
    function delete(const aString: string): Integer;
    function Hash(const S: string): Pointer;
    function HashEX(const S: string; HashValue: Integer): Pointer;
  end;

  TmwChainedBaseHashWord = class(TmwBaseHashWord)
  private
    fChain: TList;
    procedure SetChain(const Value: TList);
  protected
     property Chain: TList read fChain write SetChain;
  public
    destructor Destroy; override;
  end;

  TmwChainedGenericHashList = class(TmwGenericHashList)
  private
    fChain: TList;
    function GetCount: Integer;
    function GetEntry(Index: Integer): TmwChainedBaseHashWord;
  protected
     property Chain: TList read fChain;
  public
    constructor Create(Primary, Secondary: TmwStringHash; aCompare: TmwStringHashCompare);
    destructor Destroy; override;
    procedure Add(const aHashWord: TmwChainedBaseHashWord);
    property Count: Integer read GetCount;
    property Entry[Index: Integer]: TmwChainedBaseHashWord read GetEntry;
  end;

function CrcHash(const aString: string): integer;
function ICrcHash(const aString: string): integer;
function SmallCrcHash(const aString: string): integer;
function ISmallCrcHash(const aString: string): integer;
function TinyHash(const aString: string): Integer;
function ITinyHash(const aString: string): Integer;
function HashCompare(const Str1: string; const Str2: string): Boolean;
function IHashCompare(const Str1: string; const Str2: string): Boolean;

function HashSecondaryOne(const aString: string): Integer;
function HashSecondaryTwo(const aString: string): Integer;

procedure InitTables;

implementation

procedure InitTables;
var
  I, K: Char;
  Temp: Byte;
begin
  for I := #0 to #255 do
  begin
    mwHashTable[I] := Ord(I);
  end;
  RandSeed := 255;
  for I := #1 to #255 do
  begin
    repeat
      K := Char(Random(255));
    until K <> #0;
    Temp := mwHashTable[I];
    mwHashTable[I] := mwHashTable[K];
    mwHashTable[K] := Temp;
  end;
  for I := #0 to #255 do
    mwInsensitiveHashTable[I] := mwHashTable[AnsiLowerCase(string(I))[1]];
end;

{ based on a Hasch function by Cyrille de Brebisson }

function CrcHash(const aString: string): integer;
var
  I: Integer;
begin
  Result := 0;
  for i := 1 to length(aString) do
  begin
    Result := (Result shr 4) xor (((Result xor mwHashTable[aString[I]]) and $F) * $1000);
    Result := (Result shr 4) xor (((Result xor (ord(mwHashTable[aString[I]]) shr 4)) and $F) * $1000);
  end;
  if Result = 0 then Result := Length(aString) mod 8 + 1;
end;

function ICrcHash(const aString: string): integer;
var
  I: Integer;
begin
  Result := 0;
  for i := 1 to length(aString) do
  begin
    Result := (Result shr 4) xor (((Result xor mwInsensitiveHashTable[aString[I]]) and $F) * $1000);
    Result := (Result shr 4) xor (((Result xor (ord(mwInsensitiveHashTable[aString[I]]) shr 4)) and $F) * $1000);
  end;
  if Result = 0 then Result := Length(aString) mod 8 + 1;
end;

function SmallCrcHash(const aString: string): integer;
var
  I: Integer;
begin
  Result := 0;
  for i := 1 to length(aString) do
  begin
    Result := (Result shr 4) xor (((Result xor mwHashTable[aString[I]]) and $F) * $80);
    Result := (Result shr 4) xor (((Result xor (ord(mwHashTable[aString[I]]) shr 4)) and $F) * $80);
    if I = 3 then break;
  end;
  if Result = 0 then Result := Length(aString) mod 8 + 1;
end;

function ISmallCrcHash(const aString: string): integer;
var
  I: Integer;
begin
  Result := 0;
  for i := 1 to length(aString) do
  begin
    Result := (Result shr 4) xor (((Result xor mwInsensitiveHashTable[aString[I]]) and $F) * $80);
    Result := (Result shr 4) xor (((Result xor (ord(mwInsensitiveHashTable[aString[I]]) shr 4)) and $F) * $80);
    if I = 3 then break;
  end;
  if Result = 0 then Result := Length(aString) mod 8 + 1;
end;

function TinyHash(const aString: string): Integer;
var
  I: Integer;
begin
  Result := Length(aString);
  for i := 1 to length(aString) do
  begin
    inc(Result, mwHashTable[aString[I]]);
    Result := Result mod 128 + 1;
    if I = 2 then break;
  end;
end;

function ITinyHash(const aString: string): Integer;
var
  I: Integer;
begin
  Result := Length(aString);
  for i := 1 to length(aString) do
  begin
    inc(Result, mwInsensitiveHashTable[aString[I]]);
    Result := Result mod 128 + 1;
    if I = 2 then break;
  end;
end;

function HashCompare(const Str1: string; const Str2: string): Boolean;
var
  I: Integer;
begin
  if Length(Str1) <> Length(Str2) then
  begin
    Result := False;
    Exit;
  end;
  Result := True;
  for I := 1 to Length(Str1) do
    if Str1[I] <> Str2[I] then
    begin
      Result := False;
      Exit;
    end;
end;

function IHashCompare(const Str1: string; const Str2: string): Boolean;
var
  I: Integer;
begin
  if Length(Str1) <> Length(Str2) then
  begin
    Result := False;
    Exit;
  end;
  Result := True;
  for I := 1 to Length(Str1) do
    if mwInsensitiveHashTable[Str1[I]] <> mwInsensitiveHashTable[Str2[I]] then
    begin
      Result := False;
      Exit;
    end;
end;

function HashSecondaryOne(const aString: string): Integer;
begin
  Result := Length(aString);
  inc(Result, mwInsensitiveHashTable[aString[Length(aString)]]);
  Result := Result mod 16 + 1;
  inc(Result, mwInsensitiveHashTable[aString[1]]);
  Result := Result mod 16 + 1;
end;

function HashSecondaryTwo(const aString: string): Integer;
var
  I: Integer;
begin
  Result := Length(aString);
  for I := Length(aString) downto 1 do
  begin
    inc(Result, mwInsensitiveHashTable[aString[I]]);
    Result := Result mod 32 + 1;
  end;
end;

{ TmwBaseGenericHashList }

procedure TmwBaseGenericHashList.AdjustCapacity;
begin
  while (fCapacity > 0) and (fList[fCapacity] = nil) do
    dec(fCapacity);
  ReallocMem(FList, (fCapacity) * SizeOf(Pointer));
end;

procedure TmwBaseGenericHashList.Clear;
var
  I: Integer;
begin
  for I := 1 to fCapacity do
    if fList[I] <> nil then
      fList[I].Free;
  ReallocMem(FList, 0);
  fCapacity := 0;
end;

constructor TmwBaseGenericHashList.Create(aCompare: TmwStringHashCompare);
begin
  fCompare := aCompare;
end;

destructor TmwBaseGenericHashList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TmwBaseGenericHashList.Get(Index: Integer): Pointer;
begin
  Result := nil;
  if (Index > 0) and (Index <= fCapacity) then
    Result := fList[Index];
end;

procedure TmwBaseGenericHashList.Put(Index: Integer; Item: Pointer);
begin
  if (Index > 0) and (Index <= fCapacity) then
    fList[Index] := Item;
end;

procedure TmwBaseGenericHashList.SetCapacity(NewCapacity: Integer);
var
  I, OldCapacity: Integer;
begin
  if NewCapacity > fCapacity then
  begin
    ReallocMem(FList, (NewCapacity) * SizeOf(Pointer));
    OldCapacity := fCapacity;
    FCapacity := NewCapacity;
    for I := OldCapacity + 1 to NewCapacity do Items[I] := nil;
  end;
end;

{ TmwGenericHashStrings }

procedure TmwGenericHashStrings.Add(const aHashWord: TmwBaseHashWord);
begin
  SetCapacity(Capacity + 1);
  fList[Capacity] := aHashWord;
end;

function TmwGenericHashStrings.delete(const aString: string): Integer;
var
  I: Integer;
begin
  Result := Capacity;
  for I := 1 to Capacity do
  begin
    if fCompare(TmwBaseHashWord(fList[I]).S, aString) then
    begin
      fList[I].Free;
      if I < Capacity then
        System.Move(fList[I + 1], fList[I],
          (fCapacity - I) * SizeOf(Pointer));
      dec(fCapacity);
      ReallocMem(FList, (fCapacity) * SizeOf(Pointer));
      Result := Capacity;
      exit;
    end;
  end;
end;

{ TmwGenericHashItems }

procedure TmwGenericHashItems.Add(const aHashWord: TmwBaseHashWord);
var
  HashWord: TmwBaseHashWord;
  HashStrings: TmwGenericHashStrings;
  HashVal: Integer;
begin
  HashVal := fHash(aHashWord.S);
  SetCapacity(HashVal);
  if Items[HashVal] = nil then
  begin
    Items[HashVal] := aHashWord;
  end else
    if fList[HashVal] is TmwGenericHashStrings then
    begin
      TmwGenericHashStrings(Items[HashVal]).Add(aHashWord);
    end else
    begin
      HashWord := Items[HashVal];
      HashStrings := TmwGenericHashStrings.Create(fCompare);
      Items[HashVal] := HashStrings;
      HashStrings.Add(HashWord);
      HashStrings.Add(aHashWord)
    end;
end;

constructor TmwGenericHashItems.Create(aHash: TmwStringHash; aCompare: TmwStringHashCompare);
begin
  inherited Create(aCompare);
  fHash := aHash;
end;

function TmwGenericHashItems.delete(const aString: string): Integer;
var
  HashValue: Integer;
begin
  HashValue := fHash(aString);
  Result := Capacity;
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if fList[HashValue] is TmwBaseHashWord then
    begin
      if fCompare(TmwBaseHashWord(fList[HashValue]).S, aString) then
      begin
        fList[HashValue].Free;
        fList[HashValue] := nil;
        AdjustCapacity;
        Result := Capacity;
      end;
    end else
      if TmwGenericHashStrings(fList[HashValue]).Delete(aString) = 0 then
      begin
        fList[HashValue].Free;
        fList[HashValue] := nil;
        AdjustCapacity;
        Result := Capacity;
      end;
  end;
end;

function TmwGenericHashItems.Hash(const S: string): Pointer;
begin
  Result := HashEX(S, fHash(S));
end;

function TmwGenericHashItems.HashEX(const S: string; HashValue: Integer): Pointer;
var
  I: Integer;
begin
  Result := nil;
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if fList[HashValue] is TmwBaseHashWord then
    begin
      if fCompare(TmwBaseHashWord(fList[HashValue]).S, S) then
        Result := Items[HashValue]
    end else
      for I := 1 to TmwGenericHashStrings(fList[HashValue]).Capacity do
      begin
        if fCompare(TmwBaseHashWord(TmwGenericHashStrings(fList[HashValue])[I]).S, S) then
        begin
          Result := TmwBaseHashWord(TmwGenericHashStrings(fList[HashValue])[I]);
          exit;
        end;
      end;
  end;
end;

{ TmwGenericHashList }

constructor TmwGenericHashList.Create(Primary, Secondary: TmwStringHash; aCompare: TmwStringHashCompare);
begin
  inherited Create(aCompare);
  fHash := Primary;
  fSecondaryHash := Secondary;
end;

procedure TmwGenericHashList.Add(const aHashWord: TmwBaseHashWord);
var
  HashWord: TmwBaseHashWord;
  HashValue: Integer;
  HashItems: TmwGenericHashItems;
begin
  HashValue := fHash(aHashWord.S);
  SetCapacity(HashValue);
  if Items[HashValue] = nil then
  begin
    Items[HashValue] := aHashWord;
  end else
    if fList[HashValue] is TmwGenericHashItems then
    begin
      TmwGenericHashItems(Items[HashValue]).Add(aHashWord);
    end else
    begin
      HashWord := Items[HashValue];
      HashItems := TmwGenericHashItems.Create(fSecondaryHash, fCompare);
      Items[HashValue] := HashItems;
      HashItems.Add(HashWord);
      HashItems.Add(aHashWord);
    end;
end;

function TmwGenericHashList.delete(const aString: string): Integer;
var
  HashValue: Integer;
begin
  HashValue := fHash(aString);
  Result := Capacity;
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if fList[HashValue] is TmwBaseHashWord then
    begin
      if fCompare(TmwBaseHashWord(fList[HashValue]).S, aString) then
      begin
        fList[HashValue].Free;
        fList[HashValue] := nil;
        AdjustCapacity;
        Result := Capacity;
      end;
    end else
      if TmwGenericHashItems(fList[HashValue]).Delete(aString) = 0 then
      begin
        fList[HashValue].Free;
        fList[HashValue] := nil;
        AdjustCapacity;
        Result := Capacity;
      end;
  end;
end;

function TmwGenericHashList.Hash(const S: string): Pointer;
begin
  Result := HashEX(S, fHash(S));
end;

function TmwGenericHashList.HashEX(const S: string; HashValue: Integer): Pointer;
begin
  Result := nil;
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if fList[HashValue] is TmwBaseHashWord then
    begin
      if fCompare(TmwBaseHashWord(Items[HashValue]).S, S) then
        Result := Items[HashValue];
    end else Result := TmwGenericHashItems(fList[HashValue]).Hash(S);
  end;
end;

{ TmwChainedBaseHashWord }

destructor TmwChainedBaseHashWord.Destroy;
begin
  Chain.Remove(Self);
  inherited Destroy;
end;

procedure TmwChainedBaseHashWord.SetChain(const Value: TList);
begin
  fChain := Value;
  fChain.Add(Self);
end;

{ TmwChainedGenericHashList }

procedure TmwChainedGenericHashList.Add(
  const aHashWord: TmwChainedBaseHashWord);
begin
  aHashWord.Chain:= fChain;
  inherited Add(aHashWord);
end;

constructor TmwChainedGenericHashList.Create(Primary,
  Secondary: TmwStringHash; aCompare: TmwStringHashCompare);
begin
  inherited Create(Primary, Secondary, aCompare);
  fChain:= TList.Create;
end;

destructor TmwChainedGenericHashList.Destroy;
begin
  fChain.Free;
  inherited Destroy;
end;

function TmwChainedGenericHashList.GetCount: Integer;
begin
  Result:= fChain.Count;
end;

function TmwChainedGenericHashList.GetEntry(
  Index: Integer): TmwChainedBaseHashWord;
begin
  Result:= fChain[Index];
end;

{$R+}
initialization
  InitTables;
end.

